# -*- coding: utf-8 -*-
"""Kiwi: a tool to combine gene-set analyses with biological networks.
:author: Francesco Gatto and Leif Väremo, 2014
:email: kiwi@sysbio.se
:affiliation: Department of Biological Engineering, Chalmers, Göteborg, Sweden"""

from kiwi_module import plot